﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("MutationPerformanceAnalyzer", "An operator that analyzes the performance of mutation.")]
  [StorableClass]
  public class MutationPerformanceAnalyzer : InitializableOperator, IStatefulItem {
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<DoubleValue> QualityAfterCrossoverParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["QualityAfterCrossover"]; }
    }
    public ILookupParameter<DoubleValue> QualityAfterMutationParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["QualityAfterMutation"]; }
    }
    public ILookupParameter<IItem> PermutationBeforeMutationParameter {
      get { return (ILookupParameter<IItem>)Parameters["PermutationBeforeMutation"]; }
    }
    public ILookupParameter<IItem> PermutationAfterMutationParameter {
      get { return (ILookupParameter<IItem>)Parameters["PermutationAfterMutation"]; }
    }
    public ILookupParameter<ItemCollection<IItem>> OperatorsParameter {
      get { return (ILookupParameter<ItemCollection<IItem>>)Parameters["Operators"]; }
    }
    public IValueParameter<ISingleObjectiveSolutionSimilarityCalculator> SimilarityCalculatorParameter {
      get { return (IValueParameter<ISingleObjectiveSolutionSimilarityCalculator>)Parameters["SimilarityCalculator"]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [Storable]
    private ScatterPlotHelper diversityPlotHelper, qualityPlotHelper;
    [Storable]
    private int cnt = 0, lastGeneration = 0;

    [StorableConstructor]
    private MutationPerformanceAnalyzer(bool deserializing) : base(deserializing) { }
    private MutationPerformanceAnalyzer(MutationPerformanceAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      diversityPlotHelper = (ScatterPlotHelper)original.diversityPlotHelper.Clone(cloner);
      qualityPlotHelper = (ScatterPlotHelper)original.qualityPlotHelper.Clone(cloner);
      cnt = original.cnt;
      lastGeneration = original.lastGeneration;
    }

    public MutationPerformanceAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));

      Parameters.Add(new LookupParameter<DoubleValue>("QualityAfterCrossover", "The evaluated quality of the child solution."));
      QualityAfterCrossoverParameter.ActualName = "TSPTourLength";

      Parameters.Add(new LookupParameter<DoubleValue>("QualityAfterMutation", "The evaluated quality of the child solution."));
      QualityAfterMutationParameter.ActualName = "TSPTourLengthM";

      Parameters.Add(new LookupParameter<IItem>("PermutationBeforeMutation"));
      PermutationBeforeMutationParameter.ActualName = "TSPTourClone";

      Parameters.Add(new LookupParameter<IItem>("PermutationAfterMutation"));
      PermutationAfterMutationParameter.ActualName = "TSPTour";

      Parameters.Add(new ValueParameter<ISingleObjectiveSolutionSimilarityCalculator>("SimilarityCalculator"));
      Parameters.Add(new LookupParameter<ItemCollection<IItem>>("Operators", "The operators and items that the problem provides to the algorithms."));

      diversityPlotHelper = new ScatterPlotHelper(false, true);
      qualityPlotHelper = new ScatterPlotHelper(false, true);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MutationPerformanceAnalyzer(this, cloner);
    }

    protected override void InitializeAction() {
      if (SimilarityCalculatorParameter.Value == null) {
        SimilarityCalculatorParameter.Value = OperatorsParameter.ActualValue.OfType<ISingleObjectiveSolutionSimilarityCalculator>().FirstOrDefault();
      }
      qualityPlotHelper.InitializePlot(Results, "Mutation Quality", "Solution Index", "Absolut Quality Difference");
      diversityPlotHelper.InitializePlot(Results, "Mutation Diversity", "Solution Index", "Diversity");
      Reset();
    }

    public override IOperation Apply() {
      Initialize();
      Point2D<double> curPoint, divPoint;

      var qualityCX = QualityAfterCrossoverParameter.ActualValue.Value;
      var qualityM = QualityAfterMutationParameter.ActualValue.Value;
      var solutionBefore = PermutationBeforeMutationParameter.ActualValue;
      var solutionAfter = PermutationAfterMutationParameter.ActualValue;

      Scope permutationBeforeScope = new Scope();
      Scope permutationAfterScope = new Scope();
      permutationBeforeScope.Variables.Add(new Variable(PermutationAfterMutationParameter.ActualName, solutionBefore));
      permutationAfterScope.Variables.Add(new Variable(PermutationAfterMutationParameter.ActualName, solutionAfter));

      divPoint = new Point2D<double>(cnt, SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(permutationBeforeScope, permutationAfterScope));
      curPoint = new Point2D<double>(cnt++, qualityCX - qualityM);


      string curGenStr = GenerationsParameter.ActualValue.Value.ToString();

      qualityPlotHelper.AddPoint(curGenStr, curPoint);
      diversityPlotHelper.AddPoint(curGenStr, divPoint);

      if (GenerationsParameter.ActualValue.Value != 0) {
        if (GenerationsParameter.ActualValue.Value > lastGeneration) {
          Reset();
        }
      }

      return base.Apply();
    }

    private void Reset() {
      cnt = 0;
      lastGeneration = GenerationsParameter.ActualValue.Value;
    }

    public override void ClearState() {
      qualityPlotHelper.CleanUp();
      diversityPlotHelper.CleanUp();
    }
  }
}
